# TP 20 correction : Parcours de graphes
# BCPST1B 2023-2024
# Lycée Hoche, Versailles
# L.-C. LEFÈVRE

#%% graphe d'exemple

G_exemple = {'A': ['B', 'D'], 'B': ['C'], 'C': ['E', 'F'], 'D': ['E'], 'E': ['B'], 'F': [], 'G': ['C']}

#%% tracer un graphe

import networkx as nx
import matplotlib.pyplot as plt

G = nx.DiGraph(G_exemple)
nx.draw_networkx(G, arrowsize=25, node_size=800, font_size=8, node_color="white", edgecolors="black")
plt.show()

#%% exercice 2

def parcours_profondeur(G, x):
    if x not in vus:
        vus.append(x)
        for y in G[x]:
            if y not in vus:
                parcours_profondeur(G, y)
    else:
        return

vus = []
parcours_profondeur(G_exemple, "A")
print(vus)

#%% amélioration
# emboiter la fonction récursive dans une autre
def parcours_profondeur(G, x):
    vus = []
    def f(x):
        if x not in vus:
            vus.append(x)
            for y in G[x]:
                if y not in vus:
                    f(y)
        else:
            return
    f(x)
    return vus

print(parcours_profondeur(G_exemple, "A"))

#%% exercice 3

def existe_chemin(G, a, b):
    # est-ce que b est vu, en partant depuis a ?
    vus = parcours_profondeur(G, a)
    if b in vus:
        return True
    else:
        return False

# test
print(existe_chemin(G_exemple, "A", "F"))
print(existe_chemin(G_exemple, "A", "G"))
print(existe_chemin(G_exemple, "E", "F"))
print(existe_chemin(G_exemple, "E", "A"))

#%% exercice 4

def parcours_largeur(G, x):
    if x in vus:
        return
    else:
        vus.append(x)
        for y in G[x]:
            if y not in vus and y not in avoir:
                avoir.append(y)
        if avoir == []:
            return
        else:
            y = avoir.pop(0)
            parcours_largeur(G, y)

vus = []
avoir = []
parcours_largeur(G_exemple, "A")
print(vus)

#%% idem, amélioration

def parcours_largeur(G, x):
    vus = []
    avoir = []
    def f(x):
        if x in vus:
            return
        else:
            vus.append(x)
            for y in G[x]:
                if y not in vus and y not in avoir:
                    avoir.append(y)
            if avoir == []:
                return
            else:
                y = avoir.pop(0)
                f(y)
    f(x)
    return vus

print(parcours_largeur(G_exemple, "A"))

#%% bonus
# les longueurs viennent avec

def parcours_largeur_distances(G, x):
    vus = []
    avoir = []
    # liste des distances des sommets vus
    distances = []
    # liste des distances des sommets à voir
    avoir_distances = []
    # fonction qui doit savoir d'où elle part, et avec quelle distance de l'origine elle se trouve
    def f(x, d):
        if x in vus:
            return
        else:
            # les deux listes évoluent toujours en parallèle
            vus.append(x)
            distances.append(d)
            for y in G[x]:
                if y not in vus and y not in avoir:
                    # les sommets descendants de x sont à distance d+1
                    avoir.append(y)
                    avoir_distances.append(d+1)
            if avoir == []:
                return
            else:
                # continuer sur y, qui est à distance dy
                y = avoir.pop(0)
                dy = avoir_distances.pop(0)
                f(y, dy)
    f(x, 0)
    return (vus, distances)

print(parcours_largeur_distances(G_exemple, "A"))

#%% labyrinthes

import numpy as np
import matplotlib.pyplot as plt

lab1 = [[1,1,1,1,1,0,0,0,0,0,1],
        [1,0,0,0,1,0,1,0,1,0,1],
        [1,0,1,0,1,0,1,0,1,0,1],
        [1,0,1,0,1,0,1,0,1,0,1],
        [1,0,1,0,1,0,1,0,1,0,1],
        [0,0,1,0,0,0,1,0,1,0,0],
        [1,1,1,1,1,1,1,0,1,0,1],
        [1,0,0,0,1,0,0,0,1,0,1],
        [1,0,1,0,1,0,1,1,1,0,1],
        [1,0,1,1,1,0,1,0,0,0,1],
        [1,0,0,0,0,0,1,1,0,1,1]]

# dessine avec matplotlib
X1 = np.array(lab1)
plt.tick_params(left=False, labelleft=False, bottom=False, labelbottom=False)
plt.imshow(1-X1, cmap="gray", vmin=0, vmax=1)
plt.show()

# transforme en dictionnaire d'adjacence
def convertit(L):
    n = len(L)
    p = len(L[0])
    G = {}
    for i in range(n):
        for j in range(p):
            if L[i][j] == 0:
                G[(i, j)] = []
                if i >= 1 and L[i-1][j] == 0:
                    G[(i, j)].append((i-1, j))
                if j >= 1 and L[i][j-1] == 0:
                    G[(i, j)].append((i, j-1))
                if i <= n-2 and L[i+1][j] == 0:
                    G[(i, j)].append((i+1, j))
                if j <= p-2 and L[i][j+1] == 0:
                    G[(i, j)].append((i, j+1))
    return G

G1 = convertit(lab1)
print(G1)

#%% test en profondeur

print(parcours_profondeur(G1, (5, 0)))

#%% test en largeur

print(parcours_largeur(G1, (5, 0)))

#%% le générateur de labyrinthes
# merci à Quentin FORTIER pour ce code

import numpy as np
import matplotlib.pyplot as plt
import networkx as nx
import random

n = 10
p = 10

def generate_labyrinth(n, p):
    G = nx.Graph()
    G.add_nodes_from((i, j) for i in range(n) for j in range(p))
    visited = [[False]*p for _ in range(n)]
    def dfs(i, j):
        visited[i][j] = True
        neighbors = [(i+1, j), (i-1, j), (i, j+1), (i, j-1)]
        random.shuffle(neighbors)
        for x, y in neighbors:
            if 0 <= x < n and 0 <= y < p and not visited[x][y]:
                G.add_edge((i, j), (x, y))
                dfs(x, y)
    dfs(n - 1, 0)
    return G

G = generate_labyrinth(n, p)
L = [[1 for j in range (2*p-1)] for i in range(2*n-1)]
for i in range(n):
    for j in range(p):
        if (i, j) in nx.nodes(G):
            L[2*i][2*j] = 0
for i in range(n-1):
    for j in range(p):
        if G.has_edge((i+1,j),(i,j)):
            L[2*i+1][2*j] = 0
for i in range(n):
    for j in range(p-1):
        if G.has_edge((i,j),(i,j+1)):
            L[2*i][2*j+1] = 0

X = np.array(L)
plt.tick_params(left=False, labelleft=False, bottom=False, labelbottom=False)
plt.imshow(1-X, cmap="gray", vmin=0, vmax=1)
plt.show()

# L peut être converti (avec la fonction précédente) en dictionnaire d'adjacence

#%% exercice 7

# traduction directe de la formulation mathématique
def sommes_minimales(A):
    n = len(A)
    p = len(A[0])
    S = [[0 for j in range(p)] for i in range(n)]
    S[0][0] = A[0][0]
    for i in range(n-1):
        S[i+1][0] = S[i][0] + A[i+1][0]
    for j in range(p-1):
        S[0][j+1] = S[0][j] + A[0][j+1]
    for i in range(n-1):
        for j in range(p-1):
            if S[i+1][j] < S[i][j+1]:
                S[i+1][j+1] = S[i+1][j] + A[i+1][j+1]
            else:
                S[i+1][j+1] = S[i][j+1] + A[i+1][j+1]
    return S

# test
A = [[9, 9, 5], [1, 2, 5], [4, 6, 9]]
print(sommes_minimales(A))

#%% exercice 8.1

# on reprend
def chemin_minimal_1(A):
    n = len(A)
    p = len(A[0])
    S = [[0 for j in range(p)] for i in range(n)]
    C = [[[] for j in range(p)] for i in range(n)]
    S[0][0] = A[0][0]
    C[0][0] = [(0, 0)]
    for i in range(n-1):
        S[i+1][0] = S[i][0] + A[i+1][0]
        C[i+1][0] = C[i][0] + [(i+1, 0)]
    for j in range(p-1):
        S[0][j+1] = S[0][j] + A[0][j+1]
        C[0][j+1] = C[0][j] + [(0, j+1)]
    for i in range(n-1):
        for j in range(p-1):
            if S[i+1][j] < S[i][j+1]:
                S[i+1][j+1] = S[i+1][j] + A[i+1][j+1]
                C[i+1][j+1] = C[i+1][j] + [(i+1, j+1)]
            else:
                S[i+1][j+1] = S[i][j+1] + A[i+1][j+1]
                C[i+1][j+1] = C[i][j+1] + [(i+1, j+1)]
    # pour tester le programme, renvoyer C
    return C[n-1][p-1]

# test
A = [[9, 9, 5], [1, 2, 5], [4, 6, 9]]
print(chemin_minimal_1(A))

#%% exercice 8.2

# on remonte le chemin
def chemin_minimal_2(A):
    n = len(A)
    p = len(A[0])
    S = sommes_minimales(A)
    L = []
    # départ du coin haut-gauche
    (i, j) = (n-1, p-1)
    # tant qu'on n'est pas remonté au départ
    while i != 0 or j != 0:
        L.append((i, j))
        # on est sur la première colonne : remonter
        if j == 0: 
            i = i - 1
        # sinon, si on provenain de la gauche gauche
        elif S[i-1][j] < S[i][j-1]:
            i = i - 1
        # sinon, on arrivait d'en haut
        else:
            j = j - 1
    L.append((0, 0))
    return L

# test
A = [[9, 9, 5], [1, 2, 5], [4, 6, 9]]
print(chemin_minimal_2(A))
